# =============================
# Imports
# =============================
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

# Set Seaborn style globally
sns.set(style="whitegrid")

# =============================
# Class Definition
# =============================
class EDAHelper:
    def __init__(self, df):
        self.df = df  # Store the provided DataFrame for use in all methods

    def data_profile(self):
        profile_data = []
        for col in self.df.columns:
            profile_data.append({
                "Column": col,
                "Data Type": self.df[col].dtype,
                "Non-Null Count": self.df[col].notnull().sum(),
                "Missing Count": self.df[col].isnull().sum(),
                "Unique Values": self.df[col].nunique(),
                "Sample Values": self.df[col].dropna().unique()[:3]
            })
        return pd.DataFrame(profile_data)

    def plot_univariate(self):
        for col in self.df.columns:
            plt.figure(figsize=(6, 4))
            if pd.api.types.is_numeric_dtype(self.df[col]):
                sns.histplot(self.df[col].dropna(), kde=True)
                plt.title(f"Histogram: {col}")
            else:
                self.df[col].value_counts().plot(kind='bar')
                plt.title(f"Count Plot: {col}")
            plt.tight_layout()
            plt.show()

    def plot_bivariate(self, target):
        for col in self.df.columns:
            if col != target:
                plt.figure(figsize=(6, 4))
                if pd.api.types.is_numeric_dtype(self.df[col]):
                    sns.boxplot(x=target, y=col, data=self.df)
                else:
                    sns.countplot(x=col, hue=target, data=self.df)
                plt.title(f"{col} vs {target}")
                plt.tight_layout()
                plt.show()

    def plot_correlation_heatmap(self):
        numeric_df = self.df.select_dtypes(include='number')
        corr = numeric_df.corr()
        plt.figure(figsize=(8, 6))
        sns.heatmap(corr, annot=True, cmap='coolwarm', fmt=".2f", square=True)
        plt.title("Correlation Heatmap")
        plt.tight_layout()
        plt.show()


    def univariate_categorical(self, column, color='skyblue'):
        value_counts = self.df[column].dropna().value_counts()
        total = value_counts.sum()
        percentages = (value_counts / total * 100).round(1)
        plt.figure(figsize=(8, 4))
        bars = plt.bar(value_counts.index.astype(str), value_counts.values, color=color)
        for bar, count, pct in zip(bars, value_counts.values, percentages.values):
            plt.text(bar.get_x() + bar.get_width()/2, bar.get_height(),
                     f'{count}\n({pct}%)', ha='center', va='bottom', fontsize=10)
        plt.title(f'Categorical Distribution: {column}')
        plt.tight_layout()
        plt.show()

    def univariate_numeric(self, column, bins=20, color='skyblue'):
        data = self.df[column].dropna()
        total = len(data)
        fig, axes = plt.subplots(1, 2, figsize=(12, 4))
        counts, bin_edges, patches = axes[0].hist(data, bins=bins, color=color, edgecolor='white')
        for i in range(len(patches)):
            height = patches[i].get_height()
            if height > 0:
                percentage = f"({height / total * 100:.1f}%)"
                axes[0].text(patches[i].get_x() + patches[i].get_width() / 2,
                             height + total * 0.01,
                             f"{int(height)}\n{percentage}",
                             ha='center', va='bottom', fontsize=8)
        axes[0].set_title(f'Histogram of {column}')
        sns.kdeplot(data=data, ax=axes[1], fill=True, color=color)
        axes[1].set_title(f'KDE Plot of {column}')
        plt.tight_layout()
        plt.show()

    def categorical_vs_categorical(self, x, hue):
        data = self.df[[x, hue]].dropna()
        plt.figure(figsize=(8, 5))
        ax = sns.countplot(data=data, x=x, hue=hue, palette='Set2')
        plt.title(f"Count Plot: {x} vs {hue}")
        for container in ax.containers:
            ax.bar_label(container, fmt='%d', label_type='edge', fontsize=9, padding=3)
        plt.tight_layout()
        plt.show()

    def numeric_vs_numeric(self, x, y):
        plt.figure(figsize=(8, 5))
        sns.scatterplot(data=self.df, x=x, y=y)
        plt.title(f"Scatter Plot: {x} vs {y}")
        plt.tight_layout()
        plt.show()

    def categorical_vs_numeric(self, cat_col, num_col):
        plt.figure(figsize=(8, 5))
        sns.boxplot(data=self.df, x=cat_col, y=num_col)
        plt.title(f"Box Plot: {num_col} by {cat_col}")
        plt.tight_layout()
        plt.show()

    def multivariate_pairplot(self, columns):
        sns.pairplot(self.df[columns].dropna())
        plt.suptitle("Multivariate Pairplot", y=1.02)
        plt.show()

    def correlation_matrix(self, figsize=(8, 6), cmap='YlGnBu'):
        corr = self.df.select_dtypes(include=['float64', 'int64']).corr()
        plt.figure(figsize=figsize)
        sns.heatmap(corr, annot=True, fmt=".2f", cmap=cmap)
        plt.title("Correlation Matrix")
        plt.tight_layout()
        plt.show()

    def violin_plot(self, cat_col, num_col):
        plt.figure(figsize=(8, 5))
        sns.violinplot(data=self.df, x=cat_col, y=num_col)
        plt.title(f"Violin Plot: {num_col} by {cat_col}")
        plt.tight_layout()
        plt.show()

    def bar_mean_plot(self, cat_col, num_col):
        agg_data = self.df.groupby(cat_col)[num_col].mean().sort_values()
        plt.figure(figsize=(10, 5))
        # Assign hue explicitly to suppress deprecation warning
        sns.barplot(
            x=agg_data.index,
            y=agg_data.values,
            hue=agg_data.index,
            palette='viridis',
            legend=False
        )
        plt.ylabel(f"Average {num_col}")
        plt.title(f"Average {num_col} by {cat_col}")
        plt.xticks(rotation=45)
        plt.tight_layout()
        plt.show()

    def missing_heatmap(self):
        plt.figure(figsize=(10, 5))
        sns.heatmap(self.df.isnull(), cbar=False, cmap='Reds')
        plt.title("Missing Value Heatmap")
        plt.tight_layout()
        plt.show()

    def data_profile(self):
        profile_data = []
        for col in self.df.columns:
            profile_data.append({
                "Column": col,
                "Data Type": self.df[col].dtype,
                "Non-Null Count": self.df[col].notnull().sum(),
                "Missing Count": self.df[col].isnull().sum(),
                "Unique Values": self.df[col].nunique(),
                "Sample Values": self.df[col].dropna().unique()[:3]
            })
        return pd.DataFrame(profile_data)

    def value_counts_summary(self, column):
        data = self.df[column].value_counts(dropna=False)
        percent = (data / data.sum() * 100).round(2)
        return pd.DataFrame({'Count': data, 'Percentage': percent})


    def dtype_info(self):
        return self.df.dtypes.reset_index().rename(columns={"index": "Column", 0: "Data Type"})


    def test_churn_hypotheses_stats(dp):
        """
        Statistical testing of churn-related hypotheses using dp.df (no plots), returning results as a pandas DataFrame.
        """
        from scipy import stats
        import pandas as pd

        results = []

        # --- Hypothesis 1: Account User Count ---
        dp.df['UserCountGroup'] = dp.df['AccountUserCount'].apply(lambda x: '>3 Users' if x > 3 else '≤3 Users')
        ct1 = pd.crosstab(dp.df['UserCountGroup'], dp.df['Churn'])
        _, p1, _, _ = stats.chi2_contingency(ct1)
        results.append(['H₁: Account User Count', p1, "Reject H₀" if p1 < 0.05 else "Fail to reject H₀"])

        # --- Hypothesis 2: Support Contact Frequency ---
        ct2 = pd.crosstab(dp.df['CcContactedLYBin'], dp.df['Churn'])
        _, p2, _, _ = stats.chi2_contingency(ct2)
        results.append(['H₂: Support Contact Frequency', p2, "Reject H₀" if p2 < 0.05 else "Fail to reject H₀"])

        # --- Hypothesis 3: Payment Method ---
        ct3 = pd.crosstab(dp.df['Payment'], dp.df['Churn'])
        _, p3, _, _ = stats.chi2_contingency(ct3)
        results.append(['H₃: Payment Method', p3, "Reject H₀" if p3 < 0.05 else "Fail to reject H₀"])

        # --- Hypothesis 4: Revenue Growth (YoY) ---
        dp.df['RevGrowthYoy'] = pd.to_numeric(dp.df['RevGrowthYoy'], errors='coerce')
        dp.df['RevGrowthGroup'] = pd.cut(dp.df['RevGrowthYoy'],
                                        bins=[-float('inf'), -0.1, 0.1, float('inf')],
                                        labels=['Decline', 'Stable', 'Growth'])
        ct4 = pd.crosstab(dp.df['RevGrowthGroup'], dp.df['Churn'])
        _, p4, _, _ = stats.chi2_contingency(ct4)
        results.append(['H₄: Revenue Growth (YoY)', p4, "Reject H₀" if p4 < 0.05 else "Fail to reject H₀"])

        # --- Hypothesis 5: Loyalty and Satisfaction ---
        dp.df['LoyaltyGroup'] = dp.df['TenureBin'].astype(str) + " | " + dp.df['ServiceScore'].astype(str)
        ct5 = pd.crosstab(dp.df['LoyaltyGroup'], dp.df['Churn'])
        _, p5, _, _ = stats.chi2_contingency(ct5)
        results.append(['H₅: Loyalty & Satisfaction', p5, "Reject H₀" if p5 < 0.05 else "Fail to reject H₀"])

        # Create DataFrame
        df_results = pd.DataFrame(results, columns=['Hypothesis', 'p-value', 'Decision'])
        df_results['p-value'] = df_results['p-value'].apply(lambda x: f"{x:.2e}")

        print("\n📊 Hypothesis Statistical Test Summary:\n")
        #print(df_results)

        return df_results


    def generate_report(self):
        return {
            "Data Profile": self.data_profile(),
            "Data Types": self.dtype_info(),
            "Missing Summary": self.df.isnull().sum().reset_index(name='Missing Count').rename(columns={'index': 'Column'})
        }

if __name__ == "__main__":
    # Only for local testing
    eda = EDAHelper(processed_df)
