"""
EnsembleBuilder.py
------------------
This module contains reusable builder functions for training ensemble models
as part of the CHURNGUARD churn detection pipeline.

Supported ensemble methods:
1. Voting Classifier (Hard & Soft)
2. Stacking Classifier
3. AdaBoost Classifier
4. Bagging Classifier

Each method returns a dictionary of trained models using preselected base learners.
These models can be evaluated using the ChurnEvaluator class.
"""

# =========================
# IMPORT REQUIRED LIBRARIES
# =========================
from sklearn.ensemble import VotingClassifier, StackingClassifier, AdaBoostClassifier, BaggingClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.tree import DecisionTreeClassifier
from sklearn.naive_bayes import GaussianNB
from sklearn.neighbors import KNeighborsClassifier
from sklearn.svm import SVC
from sklearn.ensemble import RandomForestClassifier
from xgboost import XGBClassifier
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler


class EnsembleBuilder:
    def __init__(self, X_train, y_train):
        self.X_train = X_train
        self.y_train = y_train

    def build_voting_classifier(self):
        lr = LogisticRegression(max_iter=1000)
        dt = DecisionTreeClassifier()
        knn = KNeighborsClassifier()

        hard_voting = VotingClassifier(estimators=[
            ('lr', lr), ('dt', dt), ('knn', knn)
        ], voting='hard')

        soft_voting = VotingClassifier(estimators=[
            ('lr', lr), ('dt', dt), ('knn', knn)
        ], voting='soft')

        hard_voting.fit(self.X_train, self.y_train)
        soft_voting.fit(self.X_train, self.y_train)

        return [
            ({"name": "Voting_Hard"}, hard_voting),
            ({"name": "Voting_Soft"}, soft_voting)
        ]

    def build_stacking_classifier(self):
        base_learners = [
            ('dt', DecisionTreeClassifier()),
            ('knn', KNeighborsClassifier())
        ]
        final_estimator = LogisticRegression(max_iter=1000)

        stacking = StackingClassifier(
            estimators=base_learners,
            final_estimator=final_estimator,
            passthrough=False
        )

        stacking.fit(self.X_train, self.y_train)

        return [({"name": "Stacking"}, stacking)]

    def build_adaboost_classifier(self):
        ada_lr = AdaBoostClassifier(estimator=LogisticRegression(max_iter=1000), n_estimators=50)
        ada_dt = AdaBoostClassifier(estimator=DecisionTreeClassifier(), n_estimators=50)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            ada_lr.fit(self.X_train, self.y_train)
            ada_dt.fit(self.X_train, self.y_train)

        return [
            ({"name": "AdaBoost_LR"}, ada_lr),
            ({"name": "AdaBoost_DT"}, ada_dt)
        ]

    def build_bagging_classifier(self):
        bagging_dt = BaggingClassifier(base_estimator=DecisionTreeClassifier(), n_estimators=50)
        bagging_knn = BaggingClassifier(base_estimator=KNeighborsClassifier(), n_estimators=50)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            bagging_dt.fit(self.X_train, self.y_train)
            bagging_knn.fit(self.X_train, self.y_train)

        return [
            ({"name": "Bagging_DT"}, bagging_dt),
            ({"name": "Bagging_KNN"}, bagging_knn)
        ]
